package PureMessage::Resources::Location;

use strict;
use PureMessage::Resources::Config qw( %resources_config );
use PureMessage::Config qw(expand_filename);
use PureMessage::Install::Config qw(get_instopt);
use PureMessage::Version qw($PREFIX);
use PureMessage::Resources::DBH;
use ActiveState::Config qw(:path);
use Net::Domain qw( hostfqdn );
use Sophos::IP::Iface qw(interfaces);
use base qw(Exporter);

# '%web_config' has 'port' and 'ssl' attributes for on-demand indexing
# of message_body table.
use PureMessage::Manager::Config qw(%web_config);

our @EXPORT_OK = qw( %config $object %location_config);
our %config;

our $location_config; # alias that doesn't clash with other %configs
*location_config =\%config;

our $object;
our $real_PREFIX = $ENV{PREFIX} || $ENV{PMX_TEST_PREFIX} || $PREFIX; # for tests


=head1 NAME

PureMessage::Resources::Location - Encapsulates a hosts location 

=head1 SYNOPSIS

  use PureMessage::Resources::Location;
 
  my $loc = PureMessage::Resources::Location->new();
  print $loc->id; # immutable id
  print $loc->ip; # as the db sees our ip
  
  # add/update a location record in the database
  $loc->register( id => 'mx1', hostname=> 'mx1.sophos.com',ip=>'192.168.56.1');
  $loc->pong(); # signal our record as alive, also refresh our local data with
                # the database's data.
  # attempt to create a location object with the defaults taken from
  # the system rather than from the config or database
  $loc = PureMessage::Resources::Location->Discover();

=head1 DESCRIPTION

The C<PureMessage::Resources::Location> package provides a means of uniquely
identifying a server installation accross a PureMessage deployment.

=over

=item new

Returns a Location object populated from the configuration data. If there is
no configuration data available it will have undefined values for the relevent
accessors.

=item ping

Return the last pong time for this location.

=item pong 

Update the last heard from time stamp in the database and sync the data from the
db to the local config.


=item register

Add a record for this host to the db and save the information locally for this 
host. Subsequent creation of this object will read this saved information instead
of going to the database for it. Use the C<pong> method to update the locally
stored data from the db.

=item Discover

Will return a location object with the discovered information about the system. This
includes host names and ip.  This method should be used for creation of intial 
location records in the database:
  
   my $loc = PureMessage::Resources::Location->Discover();
   $loc->register();

=item Find($location_name)

Will find and return a host location object for the host that has
the matching ip, undef if none.

=item FindByIp($ip)

Will find and return a host location object for the host that has
the matching ip, undef if none.

=item Available

Will return a list of all defined locations in this profile.

=item remove

Will delete this host location from the database and the local config
file if this record is for the current host.

=item refresh_local

Will update the locally cached location settings for this host from the
database.

=item location_id

Returns the location_id for this server as registered in the database.

=item refresh_db

Will update the database to match the current state of the location object.

=item rename($new_location_name)

Will rename the location_name for this host.

=item ip

Will return the ip as contained in the location object.

=item hostname

Will return the hostname as contained in the location objec.t

=back

=head1 AUTHOR

Sophos PLC

=head1 COPYRIGHT

Copyright (C) 2003, Sophos PLC
All Rights Reserved.

=cut


my $SQL_REGISTER = <<EOF;
   INSERT INTO cfg_location (location_name,hostname,ip,manager_port,manager_protocol,pong)
        VALUES (?,?,?,?,?,now())
EOF

my $SQL_REGISTER_WITH_ID = <<EOF;
   INSERT INTO cfg_location (location_id,location_name,hostname,ip,manager_port,manager_protocol,pong)
        VALUES (?,?,?,?,?,?,now())
EOF

my $SQL_PONG = <<EOF;
   UPDATE cfg_location set pong = now()
        WHERE location_name = ?
EOF

my $SQL_GET_LOCATION_ID = <<EOF;
  SELECT location_id from cfg_location
        WHERE location_name = ?
EOF


sub get_location_id {
  my $self = shift;
  return unless $self->{location_name};

  my $sth = DBH->prepare($SQL_GET_LOCATION_ID);
  $sth->execute($self->{location_name});
  # there should only ever be one record per location_name
  my $h = $sth->fetchrow_hashref();
  return $h->{location_id} if ($h and exists($h->{location_id}));
}


sub register {
    my $self = shift;
    my $dup_host = eval { PureMessage::Resources::Location->FindByIp($self->ip); };
    die "Duplicate host registered for ip '" . $self->ip . "'\n" if $dup_host;
    
    $self->{location_name} = $self->registerable_hostname()
        unless($self->{location_name} && $self->{location_name} ne 'unknown');
    
    if( eval { PureMessage::Resources::Location->Find($self->{location_name}) } ) {
        # already a host with that name in the db, make a unique one
        $self->{location_name} = $self->{location_name} . "_" . scalar(time);
    }
    
    undef $@;

    my $status;

    if(exists($self->{location_id}) and defined($self->{location_id})) {
        $status = DBH->do($SQL_REGISTER_WITH_ID,undef,$self->{location_id},$self->location_name,$self->hostname,$self->ip,$self->manager_port,$self->manager_protocol);
    } else {
        $status = DBH->do($SQL_REGISTER,undef,$self->location_name,$self->hostname,$self->ip,$self->manager_port,$self->manager_protocol);
        if($status) {
            $self->{location_id} = $self->get_location_id();
        }
    }
 
    $self->_to_config if($status && $status > 0);
    return $status;
}


sub _to_config {
    my $self = shift;
    my $as = edit();
    my $root = $as->root();
    path_value($root,'location.hostname',$self->hostname);
    path_value($root,'location.location_name',$self->location_name);
    path_value($root,'location.ip',$self->ip);
    path_value($root,'location.location_id',$self->location_id);
    $as->save();
    refresh();    
}

sub register_local {
    my $self = shift;
    $self->_to_config();
}

sub Discover {
    my ($CLASS,%conf) = @_;
    my $self;
    $self = bless(\%conf,$CLASS);
     
    $self->{ip} = get_instopt('location_ip');
    ($self->{ip}) = get_ips() unless $self->{ip};

    $self->{hostname} = lc(hostfqdn) || 'unknown';
    $self->{pong} = undef;
    $self->{location_name} = 'unknown';
    $self->{location_id} = undef;    

    # manager_port & manager_protocol can't be null in db,
    # so give both values anyway.
    $self->{manager_port} = (exists $web_config{port})
	? $web_config{port} : 18080;
    $self->{manager_protocol} = (exists $web_config{ssl})
	? ($web_config{ssl} ? 'https' : 'http') : 'https';

    return $self;
}

my $SQL_FIND = <<EOF;
  SELECT * FROM cfg_location
        WHERE location_name = ? 
EOF

my $SQL_FIND_BY_IP = <<EOF;
  SELECT * FROM cfg_location
        WHERE ip = ?
EOF

sub Find {
    my ($CLASS,$name) = @_;
    $CLASS = ref($CLASS) || $CLASS;
    return unless $name;
    my $sth = DBH->prepare($SQL_FIND);
    $sth->execute($name);
    # there should only ever be one record per location_name
    my $h = $sth->fetchrow_hashref();
    return unless $h;
    bless($h,$CLASS); # db fields are exactly the same as what we need
}

my $SQL_LIST_ALL = <<EOF;
    SELECT location_name 
      FROM cfg_location
      ORDER BY location_name
EOF

sub Available {
    my ($CLASS,$name) = @_;
    my @locations = ();
    $CLASS = ref($CLASS) || $CLASS;
    my $sth = DBH->prepare($SQL_LIST_ALL);
    return @locations unless $sth->execute();
    while( my $h = $sth->fetchrow_hashref() ) {
      push(@locations,$h->{location_name});
    }
    return @locations;
}

sub FindByIp {
    my ($CLASS,$ip) = @_;

    return unless $ip;
    my $sth = DBH->prepare($SQL_FIND_BY_IP);
    $sth->execute($ip);
    # there should only ever be one record per ip 
    my $h = $sth->fetchrow_hashref();
    return unless $h;
    bless($h,$CLASS); # db fields are exactly the same as what we need
}

my $SQL_REMOVE = <<EOF;
   DELETE FROM cfg_location
        WHERE location_name = ?
EOF

sub remove {
    my $self = shift;

    my $status = DBH->do($SQL_REMOVE,undef,$self->location_name);
    if($status and $config{location}{location_name} eq $self->location_name) {
        my $as = edit();
        my $root = $as->root();
        path_node($root,"location", delete => 1);
        $as->save();
        refresh();
    }
    $self->{ip} = undef;
    $self->{hostname} = undef;
    $self->{location_name} = undef;
    $self->{pong} = undef;
    $self->{location_id} = undef;

    $self->{manager_port} = undef;
    $self->{manager_protocol} = undef;

    return $status;
}

sub refresh_local {
   my $self = shift;
   my $location = $self->Find($self->location_name);
   return unless $location; # actually that's pretty unexpected that we don't exist
   $self->{ip} = $location->ip;
   $self->{hostname} = $location->hostname; 
   $self->{location_id} = $location->location_id;
   $self->_to_config();
}

my $SQL_REFRESH_DB = <<EOF;
   UPDATE cfg_location set ip = ?, hostname = ?, location_id = ?, manager_port = ?, manager_protocol = ?, pong = now()
        WHERE location_name = ?
EOF

sub refresh_db {
   my $self = shift;
   if(PureMessage::Resources::Location->Find($self->location_name)) {
     my $status;
     unless(
         ($status = DBH->do(
             $SQL_REFRESH_DB,undef,$self->ip,$self->hostname,
             $self->{location_id},$self->manager_port,$self->manager_protocol,
             $self->location_name)
         ) > 0
     ) {
         $status = DBH->do(
             $SQL_REGISTER_WITH_ID,undef,$self->{location_id},
             $self->location_name,$self->hostname,$self->ip,$self->manager_port,
             $self->manager_protocol,
         );
     }
     return $status;
   } else {
     $self->register();
   }
}

my $SQL_RENAME = <<EOF;
   UPDATE cfg_location set location_name = ?
        WHERE location_name = ?
EOF


sub rename {
   my ($self,$new_location_name) = @_;
   return unless $new_location_name and $self->location_name;

   my $status = DBH->do($SQL_RENAME,undef,$new_location_name,$self->location_name);    
   if($status and $config{location}{location_name} eq $self->location_name) {
      $self->{location_name} = $new_location_name;
      $self->refresh_local();
   }
   return $status; 
}

sub ip {
    my ($self,$value) = @_;
    $self->{ip} = $value if $value;
    return $self->get('ip');
}

sub registerable_hostname {
    my $self = shift;
    my $name = $self->hostname();
    $name =~ s/\./_/g;
    return $name;
}

sub hostname {
    my ($self,$value) = @_;
    $self->{hostname} = $value if $value;
    return $self->get('hostname');
}

sub manager_port {
    my ($self,$value) = @_;
    $self->{manager_port} = $value if $value;
    return $self->get('manager_port');
}

sub manager_protocol {
    my ($self,$value) = @_;
    $self->{manager_protocol} = $value if $value;
    return $self->get('manager_protocol');
}

sub pong {
    my $self = shift;

    my $status = DBH->do($SQL_PONG,undef,$self->location_name);
    $self->refresh_local();
    return $status;
}

# read the pong timestamp for this location
sub ping {
    my $self = shift;
    my $location = $self->Find($self->location_name);
    return unless $location; # actually that's pretty unexpected that we don't exist
    return $location->{pong};
}

sub location_id {
    my ($self,$value) = @_;
    $self->{location_name} = $value if $value;
    return $self->get('location_id');
}

sub location_name {
    my ($self,$value) = @_;
    $self->{location_name} = $value if $value;
    return $self->get('location_name');
}

sub get {
    my $self = shift;
    my $key = shift;
    return unless exists $self->{$key};
    return $self->{$key};
}

sub get_ips {    
    my @ips;
    for my $if (interfaces(ipv6 => 0)) {
        $if->{ip} .= $if->{scope} ? '%' . $if->{name} : '';
        push @ips, $if->{ip};
    }
    return @ips;
}

sub new {
    my ($CLASS,%conf) = @_;
    my $self;
    $self = bless(\%conf,$CLASS);
    unless( scalar(%conf) ) { # only load info from config if there are no args given
        foreach my $k (qw (hostname ip location_name location_id) ) {
            $self->{$k} = $config{location}{$k};
        }
    }

    $self->{manager_port} = (exists $web_config{port})
	? $web_config{port} : 18080;
    $self->{manager_protocol} = (exists $web_config{ssl})
	? ($web_config{ssl} ? 'https' : 'http') : 'https';

    $self->{pong} = undef;
    return $self;
}


sub refresh() {
    require PureMessage::Config::Validator;
    
    $object = ActiveState::Config->new(strict => 1, allow_includes => 1);
    PureMessage::Resources::Location->validate($object);
    $object->parse_file("$real_PREFIX/etc/location.conf");
    %config = $object->as_hash;
    $config{prefix} = $PREFIX;
    \%config; # return this so the refresher can get it
}

refresh();

sub edit {
   my $p = ActiveState::Config->new(strict => 1, allow_includes => 1);
   PureMessage::Resources::Location->validate($p);
   $p->parse_file(expand_filename("etc/location.conf"),writable => 1);
   return $p;
}

sub validate {
    my ($pkg, $p) = @_;
    $p->typedecl('location.ip', '#str');
    $p->typedecl('location.hostname', '#str');
    $p->typedecl('location.location_name', '#str');
    $p->typedecl('location.location_id', '#num');
    $p->default('location.location_name','unknown'); #  
}


1
