package PureMessage::Resources::Profile;
use strict;
use warnings;
use base 'Exporter';
our @EXPORT_OK = qw(sync_from_db sync_to_db);

use PureMessage::Version qw($PREFIX);
use PureMessage::Resources::Manager;
use PureMessage::Resources::Store::DB;
use PureMessage::Resources::Store::Local;
use PureMessage::Resources::Log;
use PureMessage::Resources::Config qw(%resources_config);
use PureMessage::List;
use PureMessage::MultiList;
use PureMessage::Multifile::Sync;
use ActiveState::Config qw(:path);
use File::Copy;
use File::Path qw/mkpath rmtree/;
use PureMessage::Sieve qw(sieve_parse_file); 
use PureMessage::Config qw(expand_filename);

my $LOG = PureMessage::Resources::Log->new();
my %lists_in_use;
our $SIEVE_FILE = expand_filename("etc/policy.siv");

sub sync_from_db {
    my ($opts, @resources) = @_;

    # ensure that all resources we have in the db are also defined locally
    my $available = PureMessage::Resources::Manager->AvailableDB();
    foreach my $r (keys %$available) {
        my $def = PureMessage::Resources::Manager->FindDB($r);
        unless(my $a = PureMessage::Resources::Manager->FindLocal($r)) {
            # then go ahead and add it
            die "Unable to Add a definition for resource '$r' to the local store.\n"
                unless PureMessage::Resources::Manager->DefineLocal(%$def);
            print "Added resource definition for '$r' to local store\n";
            $LOG->log('INFO', "Added resource definition for '$r' to local store");
        }

    }
    my $db_rm = PureMessage::Resources::Manager->new( 
                    store => PureMessage::Resources::Store::DB->new(), 
                    writable => 1 );
    my $local_rm = PureMessage::Resources::Manager->new( 
                    store => PureMessage::Resources::Store::Local->new(), 
                    writable => 1);
    
    # ensure that all resources that have instances in the db have instances
    # locally. also remove any lists that don't appear in the db and that
    # have an attribute of resource_controlled in their args
    my %found;
    $db_rm->enumerate(sub {
                          my $r = shift;
                          my $native = $r->native() if($r->type eq 'list');
                          if (my $local = $local_rm->find($r->name)) {
                              return unless $r->type eq 'list';
                              $found{$native->id()} = $r->name();
                          } 
                          else {
                              print "Adding a resource of " . $r->name() . "\n";
                              $LOG->log('INFO','Adding a resource of ' . $r->name());
                              $local_rm->add($r->name);
                              return unless $r->type eq 'list';
                              $found{$native->id()} = $r->name();
                          }
                      }
                     );
   
    my @synced_resources;
    if (@resources) {
        $local_rm->begin_batch();
        foreach my $name (@resources) {
            my $res = $local_rm->find($name);
            unless ($res) {
                die "Unable to add resource '$name'\n" 
                    unless $res = $local_rm->add($name);
                print "Added an instance of resource '$name'.\n";
                $LOG->log('INFO',"Added an instance of resource '$name'");
            }
            my $dest = $db_rm->find($name);
            if ($dest) {
                print "syncing $name ";
                my $count = eval { $local_rm->sync($dest, $opts->{force} 
                                                       ?  (ignore_errors => 1, 
                                                           force => 1) 
                                                       : ()
                                                   ) };
                if ($@) {
                    print "ERROR: $@\n";
                    $LOG->log('INFO',"Error syncronizing resource '$name': $@");
                } 
                else {
                    print "$count changes needed\n";
                    push @synced_resources, 
                         [$res, $res->name . " ($count changes)" ];
                }
            } 
            else {
                print "Unable to find resource $name\n";
            }
        }
        $local_rm->end_batch();
    } 
    else {
        my %extra = ();
        if ($opts->{unsafe}) {
            if ($opts->{force}) {
                $extra{ignore_errors} = 1;
                $extra{force} = 1;
            }
        }
        my $sub = sub { 
            my $r = shift;
            print "sync: " . $r->name . "\n";
            my $localRes = $local_rm->find($r->name());
            unless ($localRes) {
                die "Unable to add resource '" . $r->name . "'\n" 
                    unless $localRes = $local_rm->add($r->name);
                print "Added an instance of resource '" . $r->name . "'.\n";
                $LOG->log('INFO','Added an instance of resource "' . $r->name . '"');
            }
            my $from_version = $local_rm->change_id($r->name) || 0;
            print "syncing " . $r->name . "..";
            my $count = eval { $local_rm->sync($r,%extra) };
            if ($@) {
                print "ERROR: $@\n";
                $LOG->log('INFO','Error syncronizing resource ' . $r->name . ":$@");
            } else {
                print "$count changes needed\n";
                push(@synced_resources, [ $r, $r->name . " ($count changes)" ]);
            }
        };

        if (!$opts->{all}) {
            if (ref($resources_config{'import'}) eq 'ARRAY') {
                $local_rm->begin_batch();
                foreach my $resource (@{$resources_config{'import'}}) {
                    my $res = $db_rm->find($resource);
                    next unless $res;
                    $sub->($res);
                }
                $local_rm->end_batch();
            } 
            else {
                print "No resources defined for import.\n";
                return;
            }
        } 
        else { # doing all
            _lists_in_use();
            _remove_multilists();

            $local_rm->begin_batch();
            $db_rm->enumerate($sub);
            $local_rm->end_batch();

            my @existing_lists = PureMessage::List->Available();
            my %bad_lists;
            foreach my $l (@existing_lists) {
                $bad_lists{$l->{id}} = $l->{id} 
                    if exists($l->{args}) and exists($l->{args}{resource_controlled})
                       and $l->{args}{resource_controlled} 
                       and not exists($found{$l->{id}});
            }
            
            # Don't delete in use lists
            for (keys %lists_in_use) {
                if (exists $bad_lists{$_}) {
                    $LOG->log('INFO', "Cannot delete $_ since it is in use");
                    delete $bad_lists{$_};
                }
            }

            if (keys(%bad_lists)) {
                print "Harvesting deleted lists and maps..\n";
                my $as = ActiveState::Config->new();
                my $sc = PureMessage::Config->new(writable => 1);
                $as->parse_file("$PREFIX/etc/lists.conf",writable => 1);
                foreach my $list_id (keys %bad_lists) {
                    print "deleting $list_id..\n";
                    $LOG->log('INFO', "Deleting list '$list_id'");
                    $local_rm->remove($bad_lists{$list_id});
                    unlink("$PREFIX/etc/data/lists/$list_id");
                    unlink("$PREFIX/etc/data/lists/$list_id.cdb");
                    path_nodes($as->root, "list#$list_id", delete => 1);
                    print "deleting: sync.resource#$bad_lists{$list_id}\n";
                    $LOG->log('INFO', "Removing sync.resource#$bad_lists{$list_id}");
                    path_nodes($sc->root, "sync.resource#$bad_lists{$list_id}",
                               delete => 1);
                }
                $as->save();
                $sc->save();
            }
        }
    }

    # Sync multifiles from the DB unless the user tried to sync specific
    # resources
    unless (@resources) {
        eval {
            if (PureMessage::Multifile::Sync->new->sync) {
                print "Multifiles updated from database.\n";
            }
            else {
                print "Multifiles up to date - no changes required.\n";
            }
        };
        $LOG->log('WARN', "Error while syncing multifiles: $@\n") if $@;
    }

    if(@synced_resources) {
        my $args = join(', ', map { $_->[1] } @synced_resources);
        my $location  = PureMessage::Resources::Location->new();
        $LOG->log('INFO', sprintf('location "%s" did a sync-from-db: %s',
                                  $location->location_name, $args));

        update_pristine_copies(\@synced_resources);
    }
}

sub sync_to_db {
    my ($opts, @resources) = @_;

    # ensure that all resources we have locally are also defined in the DB
    my $available = PureMessage::Resources::Manager->AvailableLocal();
    foreach my $r (keys %$available) {
        my $def = PureMessage::Resources::Manager->FindLocal($r);
        unless (my $a = PureMessage::Resources::Manager->FindDB($r)) {
            # then go ahead and add it
            die "Unable to Add a definition for resource '$r' to the DB store.\n"
                    unless PureMessage::Resources::Manager->DefineDB(%$def);
            print "Added resource definition for '$r' to db store\n";
            $LOG->log('INFO', "Added resource definition for '$r' to db store\n");
        }
    }
    my $db_rm = PureMessage::Resources::Manager->new( 
                    store => PureMessage::Resources::Store::DB->new, writable => 1 );
    my $local_rm = PureMessage::Resources::Manager->new( 
                    store => PureMessage::Resources::Store::Local->new, writable => 1);
    my @synced_resources;
    if (@resources) {
        $local_rm->begin_batch();
        foreach my $name (@resources) {
            my $res = $local_rm->find($name);
            unless ($res) {
                print "Unable to find local resource '$name'\n";
                next;
            }
            # Ensure the db resource instance exists
            unless ($db_rm->find($name)) {
                if ($db_rm->add($name)) {
                    my $msg = "Added resource '$name' to db store.";
                    print "$msg\n";
                    $LOG->log('INFO', $msg);
                } else {
                    die "Unable to add resource '$name' to db\n";
                }
            }
            print "syncing $name ";
            my $count = eval { 
                $db_rm->sync($res, $opts->{force} ? (ignore_errors => 1, force => 1) 
                                                  : () );
            };
            if ($@) {
                print "ERROR: $@\n";
                $LOG->log('INFO', "Error syncronizing $name: $@");
            } 
            else {
                print "$count changes needed\n";
                push @synced_resources, 
                     [$res, $res->name . " ($count changes)" ];
            }
        }
        $local_rm->end_batch();
    } 
    else {
        my $sub = sub { 
            my $r = shift;
            print "syncing " . $r->name . " ";
            my $count = eval {
                $db_rm->sync($r, $opts->{force}
                                 ? (ignore_errors => 1, force => 1) : ());
            };
            if ($@) {
                my $name = $r->name;
                print "ERROR: $@\n";
                $LOG->log('INFO', "Error syncronizing $name: $@");
            } 
            else {
                print "$count changes needed\n";
                push @synced_resources, 
                     [$r, $r->name . " ($count changes)" ];
            }
        };

        if (!$opts->{all}) {
            if (ref($resources_config{export}) eq 'ARRAY') {
                foreach my $resource (@{$resources_config{export}}) {
                    my $res = $local_rm->find($resource);
                    next unless $res;
                    $sub->($res);
                }
            } 
            else {
                print "No resources defined for import.\n";
                return;
            }
        } 
        else {
            $local_rm->enumerate($sub);
        }

    }

    if (@synced_resources) {
        my $location  = PureMessage::Resources::Location->new();
        my $args = join(', ', map { $_->[1] } @synced_resources);
        $LOG->log('INFO', sprintf('location "%s" did a sync-to-db: %s',
                                  $location->location_name, $args));
        update_pristine_copies(\@synced_resources);
    }
}

sub update_pristine_copies {
    my $resources = shift;

    for my $res (map { $_->[0] } @$resources) {
        next unless $res->{pristine_filename};
        (my $pris_dir = $res->{pristine_filename}) =~ s#^(.+)/.+#$1#;
        unless (-d $pris_dir) {
            mkpath $pris_dir or die "Can't mkpath $pris_dir: $!";
        }
        $res->update_pristine;
    }
}

sub _traverse_block {
    my $block = shift;
    foreach my $command ($block->commands) {
        if ($command->name eq 'if') {
            foreach my $cond ($command->args) {
                if ($cond->test) {
                    _traverse_tests($cond->test);
                }
                _traverse_block($cond);
            }
        }
    }
}

sub _traverse_tests {
    my @tests = @_;
    
    foreach my $test (@tests) {
        if ($test->name eq 'not') {
            _traverse_tests($test->args);
        }
        elsif ($test->name eq 'allof' or $test->name eq 'anyof') {
            _traverse_tests(($test->args)[0]->tests);
        }
        else {
            my @args = $test->args;
            my $last_one = $args[-1];
            foreach my $arg (@args) {
                next if $arg->type ne "tag" or $arg ne ":memberof";
                $lists_in_use{"$last_one"} = 1 if $last_one->type eq "string";
                if ($last_one->type eq "stringlist") {
                    $lists_in_use{$_} = 1 for split /\s/, "@{$last_one}";
                }
            }
        }
    }
}

sub _lists_in_use {
    my $tree = sieve_parse_file($SIEVE_FILE);
    _traverse_block($tree);
    return \%lists_in_use; # for test
}

sub _remove_multilists {
    my %local_mlist_ids = map { $_->{id} => $_->{source} }
        PureMessage::MultiList->Available;
    return unless %local_mlist_ids;


    # By default the store is DB
    my $db_mlist_ids = PureMessage::Resources::Manager->Available(bytype => 1)->{multilist};
    delete $local_mlist_ids{$_} for (@$db_mlist_ids);

    # Don't delete in use lists
    for (keys %lists_in_use) {
        if (exists $local_mlist_ids{$_}) {
            $LOG->log('INFO', "Cannot delete $_ since it is in use");
            delete $local_mlist_ids{$_};
        }
    }

    my $mlist_config = ActiveState::Config->new;
    $mlist_config->parse_file("$PREFIX/etc/multilists.conf", writable => 1);
    my $res_config = ActiveState::Config->new;
    $res_config->parse_file("$PREFIX/etc/resources.conf", writable => 1);
    my $ins_config = ActiveState::Config->new;
    $ins_config->parse_file("$PREFIX/etc/instances.conf", writable => 1);
    return unless $mlist_config && $res_config && $ins_config;

    while(my ($id, $path) = each %local_mlist_ids) {
        print "Deleting multilist and resource: $id";
        (undef, $path) = split /:/, $path;
        rmtree($path);
        path_nodes($mlist_config->root, "multilist#$id", delete => 1);
        path_nodes($res_config->root, "resource#$id", delete => 1);
        path_nodes($ins_config->root, "instance#$id", delete => 1);
        print " (done)\n";
        $LOG->log('INFO', "Deleted multilist and resource: $id");
    }

    $mlist_config->save;
    $res_config->save;
    $ins_config->save;
}

1;
